Unit Macro;
{$O+,F+,B-}
{**************************************************************************}
{ Notes > each macro can format: @macroname([Lenght][,justification])      }
{      @ - special char, what indicates macro                              }
{      macroname - name of macros                                          }
{      () - delimiters                                                     }
{      length - lenght of macro string                                     }
{      justification - ࠢ                                        }
{⪠ ଠ  (as is) : @date(),@time(),@ifname(filename.ext)     }
{**************************************************************************}

{**************************************************************************}
{* Configuration file syntax                                              *}
{* ; File description (this is comment`s! All chars after `;` are ignore) *}
{* [FILE]                                                                 *}
{* Ext=.BMP                                                               *}
{* TestSign=424D                                                          *}
{* Sign=0,2                                                               *}
{* Descr=BMP-file @wlhd(18)x@wlhd(22) pixel                               *}
{* [End]                                                                  *}
{**************************************************************************}

Interface
Uses Objects;

const MacroCh:char='@';

Type
     PSpecialFile=^TSpecialFile;
     TSpecialFile=object(TObject)
       FExt:PString;
       FSign:PString;
       FDesc:PString;
       FSOfs:integer;
       FSLen:integer;
       Constructor Init(FileExt,TestSign,Descr:String; SignOfs,SignLen:integer);
       Destructor Done; virtual;
     end;

     PSpecialFileCollection=^TSpecialFileCollection;
     TSpecialFileCollection=object(TCollection)
      Constructor Init(CfgFile:String);
      Procedure ReadSpecialFileConfig(CfgFile:String);
     end;


Function ProcessMacroStr(S,WorkFileName:string):string;
Function GetSpecialFileDescr(FileName:string; SFC:PSpecialFileCollection):string;

Implementation
Uses FaFuncs,Dos,FManage,Strings;

Constructor TSpecialFile.Init(FileExt,TestSign,Descr:String; SignOfs,SignLen:integer);
begin
 Inherited Init;
 FileExt:=Trim(UpperCase(FileExt));
 FileExt:='.'+DelCharStr(FileExt,['.']);
 if FileExt<>'' then FExt:=NewStr(FileExt) else FSign:=nil;
 TestSign:=Trim(UpperCase(TestSign));
 if TestSign<>'' then FSign:=NewStr(TestSign) else FSign:=nil;
 Descr:=Trim(Descr);
 if Descr<>'' then FDesc:=NewStr(Trim(Descr)) else FDesc:=nil;
 FSOfs:=SignOfs;
 FSLen:=SignLen;
end;

Destructor TSpecialFile.Done;
begin
 if FExt<>nil then DisposeStr(FExt);
 if FSign<>nil then DisposeStr(FSign);
 if FDesc<>nil then DisposeStr(FDesc);
 Inherited Done;
end;

Constructor TSpecialFileCollection.Init(CfgFile:String);
begin
 Inherited Init(5,5);
 if FileExist(CfgFile) then ReadSpecialFileConfig(CfgFile);
end;

Function GetSignOfs(S:String):Integer;
begin
 if S<>'' then GetSignOfs:=StrToInt(FirstWord(S,[',',';',' ']))
 else GetSignOfs:=0;
end;

Function GetSignLen(S:String):integer;
begin
 if S<>'' then GetSignLen:=Abs(StrToInt(LastWord(S,[',',';',' '])))
 else GetSignLen:=0;
end;

Function GetFlagName(S:String):String;
var Sbuf:String;
begin
 S:=Trim(S); SBuf:='';
 while (S<>'') and (S[1]<>'=') do begin Sbuf:=Sbuf+s[1]; delete(S,1,1); end;
 if S='' then GetFlagName:='' else GetFlagName:=UpperCase(Sbuf);
end;

Function DeleteFlagName(S,FlagName:String):String;
var SBuf:String;
begin
 Delete(S,1,Length(FlagName));
 while (S<>'') and (S[1]<>'=') do delete(S,1,1);
 if (S<>'') and (S[1]='=') then delete(S,1,1);
 DeleteFlagName:=S;
end;

Procedure TSpecialFileCollection.ReadSpecialFileConfig(CfgFile:String);
var F:Text;
    S,Sbuf:String;
    FlagName:String;
    eDescFlag:boolean;
    bDescFlag:boolean;
    {------------------}
    FileExt,TestSign,Descr:String;
    SignOfs,SignLen:integer;
    {------------------}
begin
 Assign(F,CfgFile);
 {$I-} Reset(F); {$I+}
 if IOResult<>0 then exit;
 eDescFlag:=False; bDescFlag:=False;
 while not eof(F) do
  begin
   readln(F,S);
   S:=Trim(DelComment(S));
   Sbuf:=S; {for file description string}
   S:=UpperCase(S);
   if S='' then continue;             {while}
   if not bDescFlag then
      if S='[FILE]' then begin
                           bDescFlag:=True;
                           FileExt:=''; TestSign:='';
                           Descr  :=''; SignOfs :=0;
                           SignLen:=0;
                           continue;
                         end;
   if bDescFlag then
    begin
     FlagName:=GetFlagName(S);
     if FlagName='EXT' then begin
                              FileExt:=DeleteFlagName(S,'EXT');
                              continue;
                             end;
     if FlagName='TESTSIGN' then begin
                                    TestSign:=DeleteFlagName(S,'TESTSIGN');
                                    continue;
                                  end;
     if FlagName='SIGN' then begin
                               S:=DeleteFlagName(S,'SIGN');
                               SignOfs:=GetSignOfs(S);
                               SignLen:=GetSignLen(S);
                               continue;
                              end;
     if FlagName='DESCR' then begin
                                 Descr:=DeleteFlagName(Trim(SBuf),'DESCR');
                                 continue;
                               end;
     if S='[END]' then eDescFlag:=True;
    end;
   if (bDescFlag and eDescFlag) then
    begin
     Insert(New(PSpecialFile,Init(FileExt,TestSign,Descr,SignOfs,SignLen)));
     bDescFlag:=False; eDescFlag:=False;
    end;
  end; {- while -}
 Close(F);
end;

Function IsSpecialFile(FileName:String; SFC:PSpecialFileCollection):boolean;
var E:String;
    Flag:boolean;
     Procedure DoSearchFile(Item:PSpecialFile); far;
     begin
      if Flag then exit;
      if (Item^.FExt<>nil) and (E=Item^.FExt^) then Flag:=True;
     end;
begin
 E:='.'+UpperCase(DelCharStr(GetFileExt(FileName),['.'])); Flag:=False;
 if SFC<>nil then SFC^.ForEach(@DoSearchFile);
 IsSpecialFile:=Flag;
end;

Function JpegInfo(FileName:String):string;
label l_error1,l_error2;
type TJPGimageData = record
                       Lngt : word;
                       Precision : byte;
                       Height : word;
                       Width  : word;
                       NumOfComps : byte;
                    end;
var JPG:file;
    B2:word;
    B1:byte;
    JPGimageData:TJPGimageData;
begin
  if not fileexist(FileName) then goto l_error1;
  FileMode:=0;   {24-09-99,BDS, ReadOnly FileMode}
  assign(JPG,FileName );
  {$I-} reset(JPG,1); {$I+}
  if IOResult<>0 then
    begin
     l_error2:
            Close(JPG);
     l_error1:
            jpeginfo:='????x????'; exit;
    end;
  BlockRead(JPG,B2,2);
  if (Lo(B2)<>$FF) and (Hi(B2)<>$D8) then goto l_error2;
  repeat
    BlockRead(JPG,B1,1,B2);
    if B1=$FF then
      begin
            BlockRead(JPG,B1,1,B2);
            if B1=$C0 then break;
      end;
  until (B2=0);
  if (B2=0) then goto l_error2;
  BlockRead(JPG,JPGimageData,SizeOf(TJpgImageData),B2);
  if B2<>SizeOf(TJpgImageData) then goto l_error2
  else with JPGimageData do jpegInfo:=IntToStr(swap(width))+'x'+IntToStr(swap(height));
  close(JPG);
end;

Function GFB(FName:string; FileOfs:longint):byte;
{Get Byte from File}
var F:File;
    B:Byte;
begin
 if not(FileExist(FName)) then exit;
 FileMode:=0; {24-09-99,BDS}
 Assign(f,FName);
 {$I-} reset(f,1); {$I+}
 if IOResult<>0 then exit;
 if FileOfs<0 then FileOfs:=FileSize(f)-FileOfs;
 Seek(f,FileOfs);
 BlockRead(f,B,SizeOf(Byte));
 close(f);
 GFB:=B;
end;

Function GFW(FName:string; FileOfs:longint):word;
{Get Word from File}
var F:File;
    W:Word;
begin
 if not(FileExist(FName)) then exit;
 FileMode:=0; {24-09-99,BDS}
 Assign(f,FName);
 {$I-} reset(f,1); {$I+}
 if IOResult<>0 then exit;
 if FileOfs<0 then FileOfs:=FileSize(f)-FileOfs;
 Seek(f,FileOfs);
 BlockRead(f,W,SizeOf(Word));
 close(f);
 GFW:=W;
end;

{WLH = GFW }
Function WHL(W:Word):word; assembler; {24-09-99,BDS}
asm
  mov ax,w
  xchg ah,al
end;

Function GFS(FName:string; FileOfs:longint; SLen:byte):string;
{get string from file (as is)}
var F:File;
    Buf:array[0..254] of char;
    s:string absolute Buf;
begin
 if not(FileExist(FName)) then exit;
 FileMode:=0; {24-09-99,BDS}
 Assign(f,FName);
 {$I-} reset(f,1); {$I+}
 if IOResult<>0 then exit;
 if FileOfs<0 then FileOfs:=FileSize(f)-FileOfs;
 Seek(f,FileOfs);
 FillChar(Buf,SizeOf(Buf),#0);
 BlockRead(f,Buf,SLen);
 close(f);
 s:=StrPas(Buf);
 GFS:=s;
end;

Function Hex2Str(HexSt:string):string;
{convert hexstring to string}
var sbuf:string[2];
    L:LongInt;
    s:string;
begin
 if (Length(HexSt) mod 2)<>0 then exit;
 s:='';
 while HexSt<>'' do
  begin
   sBuf:=HexSt; {2 }
   Delete(HexSt,1,2);
   L:=HexToInt(sBuf);
   s:=s+char(Lo(L));
  end;
 Hex2Str:=s;
end;

Function Str2Hex(s:string):string;
{convert string to hexstring}
var sbuf:string;
begin
 sbuf:='';
 while s<>'' do begin sbuf:=sbuf+IntToHex(Byte(s[1])); delete(s,1,1); end;
 Str2Hex:=sbuf;
end;

Function PrevMacro(s:string):string;
var sbuf:string;
begin
 sbuf:='';
 if (S<>'') and (s[1]=MacroCh) then PrevMacro:=sbuf
 else while (S<>'') and (s[1]<>MacroCh) do
        begin sbuf:=sbuf+s[1]; delete(s,1,1); end;
 PrevMacro:=sbuf;
end;

Procedure NextMacro(var s1,s2:string);
begin
 while (S2<>'') and (s2[1]<>MacroCh) do begin s1:=s1+s2[1]; delete(s2,1,1); end;
end;

Function GetFirstNumber(var S:String):String;
var S1:String;
    NumStr:set of char;
begin
 NumStr:=['-','+','0'..'9']; S1:='';
 while (S<>'') and not (s[1] in NumStr) do delete(s,1,1);
 if s<>'' then
  while (S<>'') and (s[1] in NumStr) do
    begin s1:=s1+s[1]; delete(s,1,1); end;
 GetFirstNumber:=S1;
end;

Procedure CutDigits(s:string;var N1,N2:longint;var Nb:byte);
{nb=0 - no digits or error, else 1 or 2 numbers}
var s1:string;
    NumStr:set of char;
begin
 Nb:=0; S:=Trim(S);
 if S='' then exit;
 S1:=GetFirstNumber(S);
 if (S1<>'') then begin N1:=StrToInt(s1); Inc(nb); end;
 if s='' then exit;
 S1:=GetFirstNumber(S);
 if (S1<>'') then begin N2:=StrToInt(s1); inc(nb); end;
end;

Procedure CutFileName(s:string; var FName:string);
begin
 FName:='';
 while ( (s[1]<>'(') and (s<>'') ) do delete(s,1,1);
 FName:=s;
 Dec(byte(FName[0]));
end;

Function CutMacro(s:string):string;
var sbuf:string;
begin
 sbuf:='';
 while (S<>'') and (s[1]<>')') do begin sbuf:=sbuf+s[1]; delete(s,1,1); end;
 delete(s,1,1);
 CutMacro:=sbuf+')';
end;

Function ChInStr(s:string;Ch:Char):byte;
var Nch:Byte;
begin
 Nch:=0;
 while Pos(Ch,s)<>0 do begin Delete(s,Pos(Ch,s),1); inc(Nch); end;
 ChInStr:=Nch;
end;

Procedure TNN(S:String; var N1,N2:Longint);
begin
 if (N1<0) or (N1>255) then N1:=Length(S);
 if not (N2 in [1..3]) then N2:=1;
end;

Function GetFileSignature(FileName:String; SignOfs,SignLen:integer):String;
var Buffer:array[1..1024] of Byte;
    F:File;
    Rb:Word;
    S:String;
    i:integer;
begin
 Assign(F,FileName);
 FileMode:=0;
 {$I-} reset(F,1); {$I+}
 if IOResult<>0 then begin GetFileSignature:=''; exit; end;
 if SignOfs<0 then Seek(F,FileSize(F)+SignOfs)
 else Seek(F,SignOfs);
 if SignLen>SizeOf(Buffer) then SignLen:=SizeOf(Buffer);
 BlockRead(F,Buffer,SignLen,Rb);
 Close(F);
 if Rb<SignLen then GetFileSignature:=''
 else begin
       S:='';
       for i:=1 to SignLen do S:=S+IntToHex(Buffer[i]);
       GetFileSignature:=S;
      end;
end;

Function ProcessMacroStr(S,WorkFileName:string):string;
var n:byte;
    sbuf:string;
    sout:string;
    sbuf2:string;
    N1,N2:longint;
    Nb:byte;
    Fn:String;
    F:text;
begin
 if (Pos(MacroCh,s)=0) or (ChInStr(s,'(')<>ChInStr(s,')')) then begin ProcessMacroStr:=s; exit; end;
 sout:=PrevMacro(s); {砫쭠  ப  }
 delete(s,Pos(sout,s),Length(sout));
 repeat
  sbuf:=CutMacro(s);  {㫨 1- }
  delete(s,Pos(sbuf,s),Length(sbuf));
  Sbuf:=UpperCase(Sbuf);
  {--------------------- ࠡ⪠ ᯥ. 䠩 -------------------------}
  if Pos(MacroCh+'WLHH',sbuf)<>0 then sout:=sout+IntToHex(gfw(WorkFileName,N1));
  if Pos(MacroCh+'WHLH',sbuf)<>0 then sout:=sout+IntToHex(whl(gfw(WorkFileName,N1)));
  if Pos(MacroCh+'WLHD',sbuf)<>0 then sout:=sout+IntToStr(gfw(WorkFileName,N1));
  if Pos(MacroCh+'WHLD',sbuf)<>0 then sout:=sout+IntToStr(whl(gfw(WorkFileName,N1)));
  if Pos(MacroCh+'BH',sbuf)<>0 then sout:=sout+IntToHex(gfb(WorkFileName,n1));
  if Pos(MacroCh+'BD',sbuf)<>0 then sout:=sout+IntToHex(gfb(WorkFileName,n1));
  if Pos(MacroCh+'GS',sbuf)<>0 then sout:=sout+gfs(WorkFileName,n1,n2);
  if Pos(MacroCh+'GSH',sbuf)<>0 then sout:=sout+Str2Hex(gfs(WorkFileName,n1,n2));
  {--------------- ⠢ ଠ樨  JPG-䠩 -------------------------}
  if Pos(MacroCh+'JPGINF',sbuf)<>0 then sout:=sout+jpeginfo(WorkFileName);{18-05-99,BDS}
                                                   {^^^^^^^^^^^^^^^^^^^^^ 10-09-99,BDS,BugFix}
  NextMacro(Sout,s);
  sbuf:=s;
 until sbuf='';
 ProcessMacroStr:=sout+s;
end;

Function GetSpecialFileDescr(FileName:string; SFC:PSpecialFileCollection):string;
var I:integer;
    Item:PSpecialFile;
    SDesc:String;
    SSign:String;
    SBuf:String;
    E:String;
begin
 SDesc:='';
 if (SFC<>nil) and IsSpecialFile(FileName,SFC) then
  begin
   E:='.'+UpperCase(DelCharStr(GetFileExt(FileName),['.']));
   for I:=0 to Pred(SFC^.Count) do
    begin
      Item:=PSpecialFile(SFC^.At(I));
      if Item^.FDesc=nil then SBuf:='' else SBuf:=Item^.FDesc^;
      if (Item^.FExt<>nil) and (E=Item^.FExt^) then
       if (Item^.FSLen<>0) then
         begin
          SSign:=GetFileSignature(FileName,Item^.FSOfs,Item^.FSLen);
          if SSign=SBuf then
           begin
             SDesc:=ProcessMacroStr(Sbuf,FileName); break;
           end;
         end
       else begin
             SDesc:=ProcessMacroStr(Sbuf,FileName); break;
           end;
    end;
  end;
 GetSpecialFileDescr:=SDesc;
end;

end.